@@ -7,6 +7,7 @@ import sys
 import time
 import traceback
 import warnings
+import subprocess
 from functools import partial
 from pathlib import Path
 
@@ -29,7 +30,7 @@ import setproctitle
 from imdb import IMDb
 from unidecode import unidecode
 
-from common import Manager, Provider, MOVIES_GROUP, PROVIDERS_PATH, SERIES_GROUP, TV_GROUP,\
+from common import Manager, Provider, Channel, MOVIES_GROUP, PROVIDERS_PATH, SERIES_GROUP, TV_GROUP,\
     async_function, idle_function
 
 
@@ -126,6 +127,7 @@ class MainWindow:
         self.icon_theme = Gtk.IconTheme.get_default()
         self.manager = Manager(self.settings)
         self.providers = []
+        self.favorite_data = []
         self.active_provider = None
         self.active_group = None
         self.active_serie = None
@@ -139,6 +141,8 @@ class MainWindow:
         self.mpv = None
         self.ia = IMDb()
 
+        self.page_is_loading = False # used to ignore signals while we set widget states
+
         self.video_properties = {}
         self.audio_properties = {}
 
@@ -179,11 +183,15 @@ class MainWindow:
             "search_button",
             "search_bar",
             "channels_box",
-            "provider_button",
-            "preferences_button",
+            "current_provider_label",
             "mpv_drawing_area",
             "stack",
+            "channel_stack",
             "fullscreen_button",
+            "mpv_top_box",
+            "mpv_bottom_box",
+            "label_channel_name",
+            "label_channel_url",
             "provider_ok_button",
             "provider_cancel_button",
             "name_entry",
@@ -204,6 +212,9 @@ class MainWindow:
             "tv_button",
             "movies_button",
             "series_button",
+            "providers_button",
+            "preferences_button",
+            "favorites_button",
             "tv_label",
             "movies_label",
             "series_label",
@@ -239,6 +250,10 @@ class MainWindow:
             "referer_entry",
             "mpv_entry",
             "mpv_link",
+            "ytdlp_local_switch",
+            "ytdlp_system_version_label",
+            "ytdlp_local_version_label",
+            "ytdlp_update_button",
             "mpv_stack",
             "spinner",
             "info_window_close_button",
@@ -250,6 +265,13 @@ class MainWindow:
             "audio_properties_label",
             "layout_properties_box",
             "layout_properties_label",
+            "favorite_button",
+            "new_channel_button",
+            "new_name_entry",
+            "new_url_entry",
+            "new_logo_entry",
+            "new_ok_button",
+            "new_cancel_button"
         ]
 
         for name in widget_names:
@@ -271,6 +293,9 @@ class MainWindow:
         self.info_window.connect("delete-event", self.on_close_info_window)
         self.info_window_close_button.connect("clicked", self.on_close_info_window_button_clicked)
 
+        self.new_ok_button.connect("clicked", self.on_new_ok_button)
+        self.new_cancel_button.connect("clicked", self.on_new_cancel_button)
+
         self.provider_ok_button.connect("clicked", self.on_provider_ok_button)
         self.provider_cancel_button.connect("clicked", self.on_provider_cancel_button)
 
@@ -278,9 +303,17 @@ class MainWindow:
         self.url_entry.connect("changed", self.toggle_ok_sensitivity)
         self.path_entry.connect("changed", self.toggle_ok_sensitivity)
 
+        self.new_name_entry.connect("changed", self.toggle_new_ok_sensitivity)
+        self.new_url_entry.connect("changed", self.toggle_new_ok_sensitivity)
+        self.new_logo_entry.connect("changed", self.toggle_new_ok_sensitivity)
+
         self.tv_button.connect("clicked", self.show_groups, TV_GROUP)
         self.movies_button.connect("clicked", self.show_groups, MOVIES_GROUP)
         self.series_button.connect("clicked", self.show_groups, SERIES_GROUP)
+        self.new_channel_button.connect("clicked", self.open_new_channel)
+        self.favorites_button.connect("clicked", self.show_favorites)
+        self.providers_button.connect("clicked", self.open_providers)
+        self.preferences_button.connect("clicked", self.open_preferences)
         self.go_back_button.connect("clicked", self.on_go_back_button)
 
         self.search_button.connect("toggled", self.on_search_button_toggled)
@@ -290,9 +323,6 @@ class MainWindow:
         self.pause_button.connect("clicked", self.on_pause_button)
         self.show_button.connect("clicked", self.on_show_button)
 
-        self.provider_button.connect("clicked", self.on_provider_button)
-        self.preferences_button.connect("clicked", self.on_preferences_button)
-
         self.new_provider_button.connect("clicked", self.on_new_provider_button)
         self.reset_providers_button.connect("clicked", self.on_reset_providers_button)
         self.delete_no_button.connect("clicked", self.on_delete_no_button)
@@ -306,11 +336,21 @@ class MainWindow:
 
         self.channels_listbox.connect("row-activated", self.on_channel_activated)
 
+        self.favorite_button.connect("toggled", self.on_favorite_button_toggled)
+
         # Settings widgets
         self.bind_setting_widget("user-agent", self.useragent_entry)
         self.bind_setting_widget("http-referer", self.referer_entry)
         self.bind_setting_widget("mpv-options", self.mpv_entry)
 
+        # ytdlp
+        self.ytdlp_local_switch.set_active(self.settings.get_boolean("use-local-ytdlp"))
+        self.ytdlp_local_switch.connect("notify::active", self.on_ytdlp_local_switch_activated)
+        self.ytdlp_system_version_label.set_text(subprocess.getoutput("/usr/bin/yt-dlp --version"))
+        if os.path.exists(os.path.expanduser("~/.cache/hypnotix/yt-dlp/yt-dlp")):
+            self.ytdlp_local_version_label.set_text(subprocess.getoutput("~/.cache/hypnotix/yt-dlp/yt-dlp --version"))
+        self.ytdlp_update_button.connect("clicked", self.update_ytdlp)
+
         # Dark mode manager
         # keep a reference to it (otherwise it gets randomly garbage collected)
         self.dark_mode_manager = XApp.DarkModeManager.new(prefer_dark_mode=True)
@@ -327,7 +367,7 @@ class MainWindow:
         item.add_accelerator("activate", accel_group, key, mod, Gtk.AccelFlags.VISIBLE)
         menu.append(item)
         self.info_menu_item = Gtk.ImageMenuItem()
-        self.info_menu_item.set_image(Gtk.Image.new_from_icon_name("dialog-information", Gtk.IconSize.MENU))
+        self.info_menu_item.set_image(Gtk.Image.new_from_icon_name("dialog-information-symbolic", Gtk.IconSize.MENU))
         self.info_menu_item.set_label(_("Stream Information"))
         self.info_menu_item.connect("activate", self.open_info)
         key, mod = Gtk.accelerator_parse("F2")
@@ -488,8 +528,18 @@ class MainWindow:
             else:
                 self.show_vod(self.active_provider.series)
 
-    def show_channels(self, channels):
-        self.navigate_to("channels_page")
+    def show_favorites(self, widget=None):
+        self.content_type = TV_GROUP
+        channels = []
+        for line in self.favorite_data:
+            info, url = line.split(":::")
+            channel = Channel(None, info)
+            channel.url = url
+            channels.append(channel)
+        self.show_channels(channels, favorites=True)
+
+    def show_channels(self, channels, favorites=False):
+        self.navigate_to("channels_page", "", favorites)
         if self.content_type == TV_GROUP:
             self.sidebar.show()
             for child in self.channels_listbox.get_children():
@@ -605,6 +655,21 @@ class MainWindow:
     def on_entry_changed(self, widget, key):
         self.settings.set_string(key, widget.get_text())
 
+    def on_ytdlp_local_switch_activated(self, widget, data=None):
+        self.settings.set_boolean("use-local-ytdlp", widget.get_active())
+        if widget.get_active():
+            self.update_ytdlp()
+
+    def update_ytdlp(self, widget=None):
+        path = os.path.expanduser("~/.cache/hypnotix/yt-dlp")
+        os.chdir(path)
+        if os.path.exists("yt-dlp"):
+            subprocess.getoutput("./yt-dlp --update")
+        else:
+            subprocess.getoutput("wget https://github.com/yt-dlp/yt-dlp/releases/latest/download/yt-dlp")
+            subprocess.getoutput("chmod a+rx ./yt-dlp")
+        self.ytdlp_local_version_label.set_text(subprocess.getoutput("~/.cache/hypnotix/yt-dlp/yt-dlp --version"))
+
     @async_function
     def download_channel_logos(self, logos_to_refresh):
         headers = {
@@ -696,36 +761,33 @@ class MainWindow:
         self.visible_search_results = 0
 
     @idle_function
-    def navigate_to(self, page, name=""):
+    def navigate_to(self, page, name="", favorites=False):
         self.go_back_button.show()
         self.search_button.show()
         self.fullscreen_button.hide()
         self.stack.set_visible_child_name(page)
         provider = self.active_provider
+        self.back_page = "landing_page"
         if page == "landing_page":
-            self.back_page = None
             self.headerbar.set_title("Hypnotix")
             if provider is None:
-                self.headerbar.set_subtitle(_("No provider selected"))
+                self.current_provider_label.set_text(_("No provider selected"))
                 self.tv_label.set_text(_("TV Channels (%d)") % 0)
                 self.movies_label.set_text(_("Movies (%d)") % 0)
                 self.series_label.set_text(_("Series (%d)") % 0)
-                self.preferences_button.set_sensitive(False)
                 self.tv_button.set_sensitive(False)
                 self.movies_button.set_sensitive(False)
                 self.series_button.set_sensitive(False)
             else:
-                self.headerbar.set_subtitle(provider.name)
+                self.current_provider_label.set_text(provider.name)
                 self.tv_label.set_text(_("TV Channels (%d)") % len(provider.channels))
                 self.movies_label.set_text(_("Movies (%d)") % len(provider.movies))
                 self.series_label.set_text(_("Series (%d)") % len(provider.series))
-                self.preferences_button.set_sensitive(True)
                 self.tv_button.set_sensitive(len(provider.channels) > 0)
                 self.movies_button.set_sensitive(len(provider.movies) > 0)
                 self.series_button.set_sensitive(len(provider.series) > 0)
             self.go_back_button.hide()
         elif page == "categories_page":
-            self.back_page = "landing_page"
             self.headerbar.set_title(provider.name)
             if self.content_type == TV_GROUP:
                 self.headerbar.set_subtitle(_("TV Channels"))
@@ -736,32 +798,33 @@ class MainWindow:
         elif page == "channels_page":
             self.fullscreen_button.show()
             self.playback_bar.hide()
-            self.headerbar.set_title(provider.name)
-            if self.content_type == TV_GROUP:
-                if self.active_group is None:
-                    self.back_page = "landing_page"
-                    self.headerbar.set_subtitle(_("TV Channels"))
-                else:
-                    self.back_page = "categories_page"
-                    self.headerbar.set_subtitle(_("TV Channels > %s") % self.active_group.name)
-            elif self.content_type == MOVIES_GROUP:
-                self.headerbar.set_subtitle(self.active_channel.name)
-                self.back_page = "vod_page"
+            if favorites:
+                self.headerbar.set_title("Hypnotix")
+                self.headerbar.set_subtitle(_("Favorites"))
             else:
-                self.headerbar.set_subtitle(self.active_channel.name)
-                self.back_page = "episodes_page"
+                self.headerbar.set_title(provider.name)
+                if self.content_type == TV_GROUP:
+                    if self.active_group is None:
+                        self.headerbar.set_subtitle(_("TV Channels"))
+                    else:
+                        self.back_page = "categories_page"
+                        self.headerbar.set_subtitle(_("TV Channels > %s") % self.active_group.name)
+                elif self.content_type == MOVIES_GROUP:
+                    self.headerbar.set_subtitle(self.active_channel.name)
+                    self.back_page = "vod_page"
+                else:
+                    self.headerbar.set_subtitle(self.active_channel.name)
+                    self.back_page = "episodes_page"
         elif page == "vod_page":
             self.headerbar.set_title(provider.name)
             if self.content_type == MOVIES_GROUP:
                 if self.active_group is None:
-                    self.back_page = "landing_page"
                     self.headerbar.set_subtitle(_("Movies"))
                 else:
                     self.back_page = "categories_page"
                     self.headerbar.set_subtitle(_("Movies > %s") % self.active_group.name)
             else:
                 if self.active_group is None:
-                    self.back_page = "landing_page"
                     self.headerbar.set_subtitle(_("Series"))
                 else:
                     self.back_page = "categories_page"
@@ -770,12 +833,13 @@ class MainWindow:
             self.back_page = "vod_page"
             self.headerbar.set_title(provider.name)
             self.headerbar.set_subtitle(self.active_serie.name)
+        elif page == "new_channel_page":
+            self.headerbar.set_title("Hypnotix")
+            self.headerbar.set_subtitle(_("New Channel"))
         elif page == "preferences_page":
-            self.back_page = "landing_page"
             self.headerbar.set_title("Hypnotix")
             self.headerbar.set_subtitle(_("Preferences"))
         elif page == "providers_page":
-            self.back_page = "landing_page"
             self.headerbar.set_title("Hypnotix")
             self.headerbar.set_subtitle(_("Providers"))
         elif page == "add_page":
@@ -803,6 +867,19 @@ class MainWindow:
         window.set_title(_("Hypnotix"))
         window.show()
 
+    def on_favorite_button_toggled(self, widget):
+        if self.page_is_loading:
+            return
+        name = self.active_channel.name
+        data = f"{self.active_channel.info}:::{self.active_channel.url}"
+        if widget.get_active() and data not in self.favorite_data:
+            print (f"Adding {name} to favorites")
+            self.favorite_data.append(data)
+        elif widget.get_active() == False and data in self.favorite_data:
+            print (f"Removing {name} from favorites")
+            self.favorite_data.remove(data)
+        self.manager.save_favorites(self.favorite_data)
+
     def on_channel_activated(self, box, widget):
         self.active_channel = widget.channel
         self.play_async(self.active_channel)
@@ -832,6 +909,7 @@ class MainWindow:
 
     @idle_function
     def before_play(self, channel):
+        self.channel_stack.set_visible_child_name("channel_page")
         self.mpv_stack.set_visible_child_name("spinner_page")
         self.video_properties.clear()
         self.video_properties[_("General")] = {}
@@ -845,6 +923,19 @@ class MainWindow:
         self.audio_bitrates.clear()
         self.spinner.start()
 
+        self.label_channel_name.set_text(channel.name)
+        self.label_channel_url.set_text(channel.url)
+
+        self.page_is_loading = True
+        data = f"{channel.info}:::{channel.url}"
+        if data in self.favorite_data:
+            self.favorite_button.set_active(True)
+            self.favorite_button.set_tooltip_text(_("Remove from favorites"))
+        else:
+            self.favorite_button.set_active(False)
+            self.favorite_button.set_tooltip_text(_("Add to favorites"))
+        self.page_is_loading = False
+
     @idle_function
     def after_play(self, channel):
         self.mpv_stack.set_visible_child_name("player_page")
@@ -1010,7 +1101,7 @@ class MainWindow:
     def on_show_button(self, widget):
         self.navigate_to("channels_page")
 
-    def on_provider_button(self, widget):
+    def open_providers(self, widget):
         self.navigate_to("providers_page")
 
     @idle_function
@@ -1092,7 +1183,15 @@ class MainWindow:
         self.init_channels_listbox()
         self.navigate_to("landing_page")
 
-    def on_preferences_button(self, widget):
+    def open_new_channel(self, widget):
+        self.new_name_entry.grab_focus()
+        self.new_name_entry.set_text("")
+        self.new_url_entry.set_text("")
+        self.new_logo_entry.set_text("")
+        self.new_ok_button.set_sensitive(False)
+        self.navigate_to("new_channel_page")
+
+    def open_preferences(self, widget):
         self.navigate_to("preferences_page")
 
     def on_new_provider_button(self, widget):
@@ -1257,6 +1356,20 @@ class MainWindow:
         provider.epg = self.epg_entry.get_text()
         self.save()
 
+    def on_new_ok_button(self, widget):
+        name = self.new_name_entry.get_text()
+        url = self.new_url_entry.get_text()
+        logo = self.new_logo_entry.get_text()
+        data = f'#EXTINF:-1 tvg-name="{name}" tvg-logo="{logo}" tvg-id="{name}" group-title="",{name}:::{url}'
+        if data not in self.favorite_data:
+            print (f"Adding {name} to favorites")
+            self.favorite_data.append(data)
+        self.manager.save_favorites(self.favorite_data)
+        self.show_favorites()
+
+    def on_new_cancel_button(self, widget):
+        self.navigate_to("landing_page")
+
     def on_provider_cancel_button(self, widget):
         self.navigate_to("providers_page")
 
@@ -1268,6 +1381,14 @@ class MainWindow:
         else:
             self.provider_ok_button.set_sensitive(True)
 
+    def toggle_new_ok_sensitivity(self, widget=None):
+        self.new_ok_button.set_sensitive(True)
+        if self.new_name_entry.get_text() == "":
+            self.new_ok_button.set_sensitive(False)
+        for widget in (self.new_url_entry, self.new_logo_entry):
+            if "://" not in widget.get_text():
+                self.new_ok_button.set_sensitive(False)
+
     def get_url(self):
         type_id = self.provider_type_combo.get_model()[self.provider_type_combo.get_active()][PROVIDER_TYPE_ID]
         if type_id == PROVIDER_TYPE_LOCAL:
@@ -1420,6 +1541,7 @@ class MainWindow:
 
     @async_function
     def reload(self, page=None, refresh=False):
+        self.favorite_data = self.manager.load_favorites()
         self.status(_("Loading providers..."))
         self.providers = []
         for provider_info in self.settings.get_strv("providers"):
@@ -1576,6 +1698,8 @@ class MainWindow:
             if self.fullscreen:
                 # Fullscreen mode
                 self.window.fullscreen()
+                self.mpv_top_box.hide()
+                self.mpv_bottom_box.hide()
                 self.sidebar.hide()
                 self.headerbar.hide()
                 self.status_label.hide()
@@ -1584,6 +1708,8 @@ class MainWindow:
             else:
                 # Normal mode
                 self.window.unfullscreen()
+                self.mpv_top_box.show()
+                self.mpv_bottom_box.hide()
                 if self.content_type == TV_GROUP:
                     self.sidebar.show()
                 self.headerbar.show()
